#define WIN32_LEAN_AND_MEAN
#define NOMINMAX
#include <windows.h>

#include <util.h>
#include <md5sums.h>
#include <soltypes_vista.h>

struct offsets_t
{
    DWORD wnd_proc;
    DWORD always_allow_bool;
    DWORD game_object;
    DWORD update_status;
};


extern "C" void* my_vista_wndproc;
extern "C" DWORD original_vista_wndproc;
extern "C" DWORD wndproc_intercept;

DWORD base;
bool always_allow_moves = false;
HWND global_hwnd = 0;
char original_title[1024];
game_object_t* game_object = 0;

// Offsets in Vista without service packs
offsets_t offsets_sp0 = { 0x33018, 0x24a39, 0x82848, 0x24cf2 };

// Offsets in Vista with service pack 1
offsets_t offsets_sp1 = { 0x32fb0, 0x24a00, 0x82848, 0x24cb9 };

// This pointer will be set by init_misc_globals().
offsets_t* offsets = 0;

//
// Returns the base of the application.
//
// Since Vista has Address Space Layout Randomization, hardcoded
// addresses is of no use. All addresses are stored as offsets from
// the start of the image, and calculated runtime.
//
void get_base()
{
    base = (DWORD)GetModuleHandle(NULL);
}

//
// Initialize global variables
//
// This also returns true if the version of the executable is
// unknown. The version is determined by calculating the md5 of the
// executable and comparing it against two known md5 sums.

// 
bool init_misc_globals()
{
    unsigned char md5[16];
    if( !get_md5_of_process( GetCurrentProcessId(), md5 ) )
	return false;

    if( !memcmp(md5, vista_sp0_md5, sizeof(md5) ) )
	offsets = &offsets_sp0;
    else if( !memcmp(md5, vista_sp1_md5, sizeof(md5) ) )
	offsets = &offsets_sp1;
    else
	return false;
    
    game_object = *(game_object_t**)(base + offsets->game_object);
    return true;
}

//
// Gets the original solitaire window title
//
void get_original_window_title()
{
    if( !global_hwnd )
	return;
    GetWindowText( global_hwnd, original_title, sizeof(original_title));
}

//
// Set always allow move on or off
//
// This is simply done by toggling setting a bool in the move
// validation function to always true. The validation function already
// has some functionality to allow any move, but it is disabled.
//
void set_always_allow_move( bool always_allow )
{
    unsigned char* ptr = (unsigned char*)(base + offsets->always_allow_bool);
    DWORD old_protect;
    if( !VirtualProtect( (void*)ptr, 1, PAGE_EXECUTE_READWRITE, &old_protect ) )
    {
	fprintf(stderr,"error setting PAGE_EXECUTE_READWRITE\n");
	return;
    }
    *ptr = always_allow ? 1 : 0;
}

//
// Calls the solitaire update status-function, which updates the time
// used and the score in the bottom of the window
//
// The function is a member function in the SolitaireGame-class, and expects
// the pointer to the object in ecx.
// 
void update_status()
{
    DWORD func = base + offsets->update_status;
    _asm
    {
        mov eax, func
	mov ecx, game_object
        call eax	    
    }
}

//
// Decrease the time used by 60 seconds
// 
void decrease_time()
{
    game_object->game_state->time -= 60;
    update_status();
}

//
// Increase the score by 10 
//
void increase_score()
{
    game_object->game_state->standard_score += 10;
    game_object->game_state->vegas_score += 10;
    game_object->game_state->untimed_score += 10;
    update_status();
}

//
// Sets the solitaire window title
// 
void set_window_title()
{
    if( !global_hwnd )
	return;
    char title[1024];
    sprintf_s(
	title,
	sizeof(title),
	"%s - TRAINER BY VALiUM!DESIGN - %c",
	original_title,
	always_allow_moves ? 'A' : 'a' );
    SetWindowText( global_hwnd, title);
}

//
// The replacement window proc.
//
// This function is called before the original window proc.
//
// The first time this function is called, it sets the global_hwnd
// variable and updates the window title.
//
//
extern "C" void WINAPI my_wndproc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam )
{
    if( !global_hwnd )
    {
	global_hwnd = hwnd;
	get_original_window_title();
	set_window_title();
    }
    if( msg == WM_KEYDOWN )
    {
	switch( wparam )
	{
	case 'A':
	    always_allow_moves = !always_allow_moves;
	    set_always_allow_move( always_allow_moves );
	    set_window_title();
	    break;
	case 'S':
	    increase_score();
	    break;
	case 'T':
	    decrease_time();
	    break;
#ifdef DEBUG	    
	case 'P':
	    log("%u %u\n", game_object->game_state->time, game_object->game_state->standard_score );
	    break;
#endif	    
	}
    }
}

//
// Patches the original window proc in the solitaire application.
//
// This is done by using the hotpatch-function prefix in the original function.
//
// The function starts with these 3 instructions:
//
//     mov edi, edi
//     push ebp
//     mov ebp, esp
//
// These instructions uses bytes, and can be replaced by long jmp
//
bool patch_wndproc()
{
    original_vista_wndproc = base + offsets->wnd_proc;
    DWORD old_protect;
    if( !VirtualProtect( (void*)original_vista_wndproc, 5, PAGE_EXECUTE_READWRITE, &old_protect ) )
    {
	fprintf(stderr,"error setting PAGE_EXECUTE_READWRITE\n");
	return false;
    }
    my_vista_wndproc = &my_wndproc;
    unsigned char* inst = (unsigned char*)original_vista_wndproc;
    *inst = 0xe9;
    
    DWORD jump_offset = (DWORD)&wndproc_intercept - (original_vista_wndproc+5);
    DWORD* jmp_ofs = (DWORD*)(inst+1);
    *jmp_ofs = jump_offset;
    return true;
}


BOOL WINAPI DllMain( HINSTANCE instance, DWORD reason, LPVOID reserved )
{
    if( reason == DLL_PROCESS_ATTACH )
    {
#ifdef DEBUG	
	init_console();
#endif
	get_base();
	if( !init_misc_globals() )
	{
#ifdef DEBUG	    
	    log("error initing stuff\n");
#else
	    MessageBox(NULL, "Unknown solitaire version", "Solitaire Trainer", MB_OK | MB_ICONERROR);
#endif
	    return TRUE;
	}
	if( !patch_wndproc() )
	{
#ifdef DEBUG	    
	    log("error patching wndproc\n");
#else
	    MessageBox(NULL, "Error patching wndproc", "Solitaire Trainer", MB_OK | MB_ICONERROR );
#endif	    
	    return TRUE;
	}
    }
    return TRUE;
}
    
