#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <soltypes_xp.h>
#include <util.h>
#include <string>
#include <vector>
#include <list>
#include <time.h>

//
// Useful adAddresses in the solitaire application 
// 
const DWORD GLOBAL_DATA_ADDRESS = 0x01007170;
const DWORD HWND_ADDRESS = 0x0100733C;
const DWORD JUMP_INIT_ADDRESS = 0x0100289b;
const DWORD INIT_ADDRESS = 0x01005581;
const DWORD CHECK_MOVE_ADDRESS = 0x010042c7;

std::string next_message; // the next message to be printed (when the score potentially gets updated)
unsigned last_score = 0; // the last score that was printed
handle_stuff_t old_func; // the old function address (the one being intercepted)
bool always_allow_moves = false;
char original_title[1024];

struct undo_t
{
    std::vector< card_t > stacks[13];
    DWORD score;
};

std::list< undo_t > undo_list;
    
const char* get_stack_str( DWORD stack )
{
    char *str[13] = { "dh", "ds", "t1", "t2", "t3", "t4", "s1", "s2","s3","s4","s5","s6","s7" };
    if( stack < 13 )
	return str[stack];
    else
	return "ERROR";
}

const char* get_card_str(card_t* card)
{
    const char* lut[] = {
	"Ac", "Ad", "Ah", "As",
	"2c", "2d", "2h", "2s",
	"3c", "3d", "3h", "3s",
	"4c", "4d", "4h", "4s",
	"5c", "5d", "5h", "5s",
	"6c", "6d", "6h", "6s",
	"7c", "7d", "7h", "7s",
	"8c", "8d", "8h", "8s",
	"9c", "9d", "9h", "9s",
	"Tc", "Td", "Th", "Ts",
	"Jc", "Jd", "Jh", "Js",
	"Qc", "Qd", "Qh", "Qs",
	"Kc", "Kd", "Kh", "Ks" };
    unsigned t = card->type & 0xff;
    if( t < 52 )
	return lut[t];
    else
	return "EE";
}

//
// Gets the original solitaire window title
//
void get_original_window_title()
{
    GetWindowText( *(HWND*)HWND_ADDRESS, original_title, sizeof(original_title));
}

//
// Sets the solitaire window title
// 
void set_window_title()
{
    char title[1024];
    sprintf_s(
	title,
	sizeof(title),
	"%s - TRAINER BY VALiUM!DESIGN - %c",
	original_title,
	always_allow_moves ? 'A' : 'a' );
    SetWindowText( *(HWND*)HWND_ADDRESS, title);
}

//
// Repaints the solitaire window
//
void repaint_window()
{
    InvalidateRect( *(HWND*)HWND_ADDRESS, NULL, 0 );
}

//
// Copies the game state to the undo buffer
//
void copy_to_undo_buf( global_data_t* data )
{
    undo_t undo_buf;
    for( unsigned i = 0; i < 13; i++ )
    {
	stack_t* stack = data->stacks[i];
	undo_buf.stacks[i].clear();
	for( unsigned j = 0; j < stack->size; j++ )
	    undo_buf.stacks[i].push_back( stack->cards[j] );
    }
    undo_buf.score = data->score;
    undo_list.push_back( undo_buf );
}

//
// Copies the newest data from the undo buffer to the game
// 
void copy_from_undo_buf( global_data_t* data )
{
    if( undo_list.empty() )
	return;

    undo_t undo_buf = undo_list.back();
    undo_list.pop_back();
    for( unsigned i = 0; i < 13; i++ )
    {
	stack_t* stack = data->stacks[i];
	for( unsigned j = 0; j < undo_buf.stacks[i].size(); j++ )
	    stack->cards[j] = undo_buf.stacks[i][j];
	stack->size = undo_buf.stacks[i].size();
    }
    data->score = undo_buf.score;
    repaint_window();
}

//
// Set always allow move on or off
//
// This is done by bypassing some of the validation checks. The
// instruction at CHECK_MOVE_ADDRESS is originally and ecx, edx (23
// ca), but is patched to jmp short $+1a (eb 1a).
//
void set_always_allow_move(bool allow)
{
    WORD* ptr = (WORD*)(CHECK_MOVE_ADDRESS);
    DWORD old_protect;
    if( !VirtualProtect( ptr, 4, PAGE_EXECUTE_READWRITE, &old_protect ) )
    {
	fprintf(stderr,"error setting PAGE_EXECUTE_READWRITE\n");
	return;
    }

    *ptr = allow ? 0x1aeb : 0xca23;
}

//
// Moves all the cards to the 4 stacks at the top, except for the last
// king
//
void cheat( global_data_t* data )
{
    // clear stacks
    for( unsigned i = 0; i < 13; i++ )
	data->stacks[i]->size = 0;

    // fill almost everything
    for( unsigned i = 0; i < 51; i++ )
	data->stacks[(i%4)+2]->cards[i/4].type = i | 0x8000;

    data->stacks[6]->cards[0].type = 51 | 0x8000;
    // set sizes
    data->stacks[2]->size = 13;
    data->stacks[3]->size = 13;
    data->stacks[4]->size = 13;
    data->stacks[5]->size = 12;
    data->stacks[6]->size = 1;
    data->time = 30*4;
    repaint_window();
}

// 
// Flips all cards face up
//
// The most significant bit in the card type is a visiblity flag. This
// function just sets the flag for all cards.
// 
void peek( global_data_t* data )
{
    for( unsigned i = 0; i < 13; i++ )
    {
	stack_t* stack = data->stacks[i];
	for( unsigned j = 0; j < stack->size; j++ )
	{
	    stack->cards[j].type |= 0x8000;
	}
    }
    repaint_window();
}

//
// Dump all stacks, useful for debug purposes
//
void dump_stacks( global_data_t* data )
{
    for( unsigned i = 0; i < 13; i++ )
    {
	stack_t* stack = data->stacks[i];
	log("%s:", get_stack_str(i) );
	for( unsigned j = 0; j < stack->size; j++ )
	    log(" %s", get_card_str( &stack->cards[j]) );
	log("\n");
    }
}

//
// Handles keypresses
//
void handle_keypress( global_data_t* data, DWORD b, DWORD c )
{
    switch( b )
    {
#ifdef DEBUG	
    case 'D':
	dump_stacks(data);
	break;
#endif	
    case 'U':
	copy_from_undo_buf(data);
	break;
    case 'C':
	cheat(data);
	break;
    case 'S':
	data->score += 10;
	break;
    case 'T':
	data->time -= 30*4;
	break;
    case 'P':
	peek(data);
	break;
    case 'A':
	always_allow_moves = !always_allow_moves;
	set_always_allow_move( always_allow_moves );
	break;
    }
    set_window_title();
}

//
// Handles solitaire events
//
void pre_handle_event( global_data_t* data, DWORD cmd, DWORD b, DWORD c )
{
    switch( cmd )
    {
    case CMD_MOVE_CARDS:
	copy_to_undo_buf(data);
	return;
    case CMD_REVEAL_CARD:
	copy_to_undo_buf(data);
	return;
    case CMD_KEYPRESS:
	handle_keypress(data, b, c);
	return;
    case CMD_MOVE_CARDS_2:
	repaint_window();
	return;
    default:
#ifdef DEBUG	
	log("%02x pre  score = %u b = %08x c = %08x\n", cmd, data->score, b, c);
#endif	
	return;
    }
}

//
// The event intercept function
//
// This function is called whenever there is an internal solitaire
// event. It takes 4 parameters:
//     data - A pointer to the global data
//     cmd - The command (event type). See soltypes_xp.h for a list of
//           commands.
//     b - command-specific parameter
//     c - command-specific parameter
//
// Whenever this function is called, the pre_handle_event() is called
// before the original solitaire function. If it had been necessary, a
// post_handle_event() could've been called after the call to the
// original function.
//
DWORD __stdcall event_intercept_func( global_data_t* data, DWORD cmd, DWORD b, DWORD c )
{
    pre_handle_event( data, cmd, b, c );
    return (*old_func)(data, cmd, b, c);
}

//
// Initialize the event intercept function
//
// Solitaire has a global function pointer to the event function. It
// is trivial to overwrite this pointer with a pointer to our own
// function.
//
void init_event_intercept()
{
    global_data_t* ptr = *(global_data_t**)GLOBAL_DATA_ADDRESS;
    old_func = ptr->func;
    ptr->func = &event_intercept_func;
}

//
// The initfunc intercept
//
// This function is called whenever solitaire initializes the internal
// datastructures. This happens whenever the user changes the game
// type and stuff like that. The other intercept 
// 
void initfunc_intercept()
{
    init_func_t func = (init_func_t)INIT_ADDRESS;
    (*func)();
    init_event_intercept();
}

//
// Initialize the initfunc intercept
//
// The solitaire initfunc starts with a jmp long addr, which is a 5
// byte instruction encoded as:
//     09 xx xx xx xx
// where xx xx xx xx is the relative offset of the jump target.
//
// The offset is relative to the next instruction, so when patching
// this instruction, the difference between my function and
// the instruction right after the jump has to be calculated.
// 
void init_initfunc_intercept()
{
    DWORD* ptr = (DWORD*)(JUMP_INIT_ADDRESS+1);
    // enable writing
    DWORD old_protect;
    if( !VirtualProtect( ptr, 4, PAGE_EXECUTE_READWRITE, &old_protect ) )
    {
	fprintf(stderr,"error patching initfunc jump\n");
	return;
    }
	
    DWORD jmp_ofs = ((DWORD)&initfunc_intercept) - (JUMP_INIT_ADDRESS+5);
    *ptr = jmp_ofs;
}

// 
// Checks if this is a supported version of solitaire
//
// This is simply done by checking if the executable is at address
// 0x01000000 and if an address in the code contains a particular
// value.
//
bool check_version()
{
    if( (DWORD)GetModuleHandle(NULL) != 0x01000000 ||
	*(DWORD*)JUMP_INIT_ADDRESS != 0x002ce1e9 )
	return false;
    return true;
}

BOOL WINAPI DllMain( HINSTANCE instance, DWORD reason, LPVOID reserved )
{
    if( reason == DLL_PROCESS_ATTACH )
    {
#ifdef DEBUG	
	init_console();
#endif
	if( !check_version() )
	{
#ifdef DEBUG	    
	    log("wrong version\n");
#else
	    MessageBox(NULL, "Unknown solitaire version", "Solitaire Trainer", MB_OK | MB_ICONERROR);
#endif	    
	    return TRUE;
	}
	
	init_event_intercept();
	init_initfunc_intercept();
	get_original_window_title();
	set_window_title();
    }
    return TRUE;
}
    
